@@ -4,7 +4,7 @@ Project: pyautd
 Created Date: 11/02/2020
 Author: Shun Suzuki
 -----
-Last Modified: 10/07/2021
+Last Modified: 14/09/2021
 Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
 -----
 Copyright (c) 2020 Hapis Lab. All rights reserved.
@@ -17,7 +17,7 @@ import math
 import numpy as np
 from functools import singledispatch
 
-from .native_methods import Nativemethods
+from .native_methods import Nativemethods, ErrorHandlerFunc, GainMode
 
 NATIVE_METHODDS = Nativemethods()
 
@@ -29,12 +29,12 @@ class Gain:
     def __del__(self):
         NATIVE_METHODDS.autd3capi.AUTDDeleteGain(self.gain_ptr)
 
-    @staticmethod
+    @ staticmethod
     def adjust_amp(amp):
         d = math.asin(amp) / math.pi
         return int(510.0 * d)
 
-    @staticmethod
+    @ staticmethod
     def grouped(gain_pairs):
         gain = Gain()
         NATIVE_METHODDS.autd3capi.AUTDGainGrouped(byref(gain.gain_ptr))
@@ -42,38 +42,38 @@ class Gain:
             NATIVE_METHODDS.autd3capi.AUTDGainGroupedAdd(gain.gain_ptr, id, gp.gain_ptr)
         return gain
 
-    @singledispatch
+    @ singledispatch
     def focal_point(pos, duty: int = 255):
         gain = Gain()
         NATIVE_METHODDS.autd3capi.AUTDGainFocalPoint(byref(gain.gain_ptr), pos[0], pos[1], pos[2], duty)
         return gain
 
-    @focal_point.register
+    @ focal_point.register
     def _(pos, amp: float = 1.0):
         return Gain.focal_point(pos, Gain.adjust_amp(amp))
 
-    @singledispatch
+    @ singledispatch
     def bessel_beam(pos, dir, theta_z, duty: int = 255):
         gain = Gain()
         NATIVE_METHODDS.autd3capi.AUTDGainBesselBeam(byref(gain.gain_ptr), pos[0], pos[1], pos[2], dir[0], dir[1], dir[2], theta_z, duty)
         return gain
 
-    @bessel_beam.register
+    @ bessel_beam.register
     def _(pos, dir, theta_z, amp: float = 1.0):
         return Gain.bessel_beam(pos, dir, theta_z, Gain.adjust_amp(amp))
 
-    @singledispatch
+    @ singledispatch
     def plane_wave(pos, dir, duty: int = 255):
         gain = Gain()
         NATIVE_METHODDS.autd3capi.AUTDGainPlaneWave(byref(gain.gain_ptr), pos[0], pos[1], pos[2], dir[0], dir[1], dir[2], duty)
         return gain
 
-    @plane_wave.register
+    @ plane_wave.register
     def _(pos, dir, amp: float = 1.0):
         duty = Gain.adjust_amp(amp)
         return Gain.plane_wave(pos, dir, duty)
 
-    @staticmethod
+    @ staticmethod
     def custom(data):
         size = len(data)
         data = np.array(data).astype(np.uint16)
@@ -83,7 +83,7 @@ class Gain:
         NATIVE_METHODDS.autd3capi.AUTDGainCustom(byref(gain.gain_ptr), data, size)
         return gain
 
-    @staticmethod
+    @ staticmethod
     def __pack_foci(foci):
         size = len(foci)
         foci_array = np.zeros([size * 3]).astype(np.float64)
@@ -99,7 +99,7 @@ class Gain:
         amps = np.ctypeslib.as_ctypes(amps)
         return amps
 
-    @staticmethod
+    @ staticmethod
     def holo_sdp(foci, amps, alpha: float = 1e-3, lambda_: float = 0.9, repeat: int = 100, normalize: bool = False):
         size = len(foci)
         foci_array = Gain.__pack_foci(foci)
@@ -112,7 +112,7 @@ class Gain:
         NATIVE_METHODDS.autd3capi_holo_gain.AUTDGainHoloSDP(byref(gain.gain_ptr), backend, foci_array, amps, size, alpha, lambda_, repeat, normalize)
         return gain
 
-    @staticmethod
+    @ staticmethod
     def holo_evd(foci, amps, gamma: float = 1, normalize: bool = False):
         size = len(foci)
         foci_array = Gain.__pack_foci(foci)
@@ -125,7 +125,7 @@ class Gain:
         NATIVE_METHODDS.autd3capi_holo_gain.AUTDGainHoloEVD(byref(gain.gain_ptr), backend, foci_array, amps, size, gamma, normalize)
         return gain
 
-    @staticmethod
+    @ staticmethod
     def holo_gs(foci, amps, repeat: int = 100):
         size = len(foci)
         foci_array = Gain.__pack_foci(foci)
@@ -138,7 +138,7 @@ class Gain:
         NATIVE_METHODDS.autd3capi_holo_gain.AUTDGainHoloGS(byref(gain.gain_ptr), backend, foci_array, amps, size, repeat)
         return gain
 
-    @staticmethod
+    @ staticmethod
     def holo_gspat(foci, amps, repeat: int = 100):
         size = len(foci)
         foci_array = Gain.__pack_foci(foci)
@@ -151,7 +151,7 @@ class Gain:
         NATIVE_METHODDS.autd3capi_holo_gain.AUTDGainHoloGSPAT(byref(gain.gain_ptr), backend, foci_array, amps, size, repeat)
         return gain
 
-    @staticmethod
+    @ staticmethod
     def holo_naive(foci, amps):
         size = len(foci)
         foci_array = Gain.__pack_foci(foci)
@@ -164,7 +164,7 @@ class Gain:
         NATIVE_METHODDS.autd3capi_holo_gain.AUTDGainHoloNaive(byref(gain.gain_ptr), backend, foci_array, amps, size)
         return gain
 
-    @staticmethod
+    @ staticmethod
     def holo_lm(foci, amps, eps1: float = 1e-8, eps2: float = 1e-8, tau: float = 1e-3, k_max: int = 5, initial=None):
         size = len(foci)
         foci_array = Gain.__pack_foci(foci)
@@ -189,7 +189,77 @@ class Gain:
             0 if initial is None else len(initial))
         return gain
 
-    @staticmethod
+    @ staticmethod
+    def holo_gauss_newton(foci, amps, eps1: float = 1e-6, eps2: float = 1e-6, k_max: int = 500, initial=None):
+        size = len(foci)
+        foci_array = Gain.__pack_foci(foci)
+        amps = Gain.__pack_amps(amps)
+
+        backend = c_void_p()
+        NATIVE_METHODDS.autd3capi_holo_gain.AUTDEigen3Backend(byref(backend))
+
+        gain = Gain()
+        NATIVE_METHODDS.autd3capi_holo_gain.AUTDGainHoloGaussNewton(
+            byref(
+                gain.gain_ptr),
+            backend,
+            foci_array,
+            amps,
+            size,
+            eps1,
+            eps2,
+            k_max,
+            initial,
+            0 if initial is None else len(initial))
+        return gain
+
+    @ staticmethod
+    def holo_gradient_descent(foci, amps, eps: float = 1e-6, step: float = 0.5, k_max: int = 2000, initial=None):
+        size = len(foci)
+        foci_array = Gain.__pack_foci(foci)
+        amps = Gain.__pack_amps(amps)
+
+        backend = c_void_p()
+        NATIVE_METHODDS.autd3capi_holo_gain.AUTDEigen3Backend(byref(backend))
+
+        gain = Gain()
+        NATIVE_METHODDS.autd3capi_holo_gain.AUTDGainHoloGradientDescent(
+            byref(
+                gain.gain_ptr),
+            backend,
+            foci_array,
+            amps,
+            size,
+            eps,
+            step,
+            k_max,
+            initial,
+            0 if initial is None else len(initial))
+        return gain
+
+    @ staticmethod
+    def holo_apo(foci, amps, eps: float = 1e-8, lambda_: float = 1.0, k_max: int = 200):
+        size = len(foci)
+        foci_array = Gain.__pack_foci(foci)
+        amps = Gain.__pack_amps(amps)
+
+        backend = c_void_p()
+        NATIVE_METHODDS.autd3capi_holo_gain.AUTDEigen3Backend(byref(backend))
+
+        gain = Gain()
+        NATIVE_METHODDS.autd3capi_holo_gain.AUTDGainHoloAPO(
+            byref(
+                gain.gain_ptr),
+            backend,
+            foci_array,
+            amps,
+            size,
+            eps,
+            lambda_,
+            k_max)
+        return gain
+
+    @ staticmethod
     def holo_greedy(foci, amps, phase_div=16):
         size = len(foci)
         foci_array = Gain.__pack_foci(foci)
@@ -199,13 +269,13 @@ class Gain:
         NATIVE_METHODDS.autd3capi_holo_gain.AUTDGainHoloGreedy(byref(gain.gain_ptr), foci_array, amps, size, phase_div)
         return gain
 
-    @staticmethod
+    @ staticmethod
     def transducer_test(idx: int, duty: int, phase: int):
         gain = Gain()
         NATIVE_METHODDS.autd3capi.AUTDGainTransducerTest(byref(gain.gain_ptr), idx, duty, phase)
         return gain
 
-    @staticmethod
+    @ staticmethod
     def null():
         gain = Gain()
         NATIVE_METHODDS.autd3capi.AUTDGainNull(byref(gain.gain_ptr))
@@ -219,13 +289,13 @@ class Modulation:
     def __del__(self):
         NATIVE_METHODDS.autd3capi.AUTDDeleteModulation(self.modulation_ptr)
 
-    @staticmethod
+    @ staticmethod
     def static(amp=255):
         mod = Modulation()
         NATIVE_METHODDS.autd3capi.AUTDModulationStatic(byref(mod.modulation_ptr), amp)
         return mod
 
-    @staticmethod
+    @ staticmethod
     def custom(data):
         size = len(data)
         data = np.array(data).astype(np.uint8)
@@ -235,7 +305,7 @@ class Modulation:
         NATIVE_METHODDS.autd3capi.AUTDModulationCustom(byref(mod.modulation_ptr), data, size)
         return mod
 
-    @staticmethod
+    @ staticmethod
     def sine(freq: int, amp=1.0, offset=0.5):
         mod = Modulation()
         NATIVE_METHODDS.autd3capi.AUTDModulationSine(byref(mod.modulation_ptr), freq, amp, offset)
@@ -247,7 +317,13 @@ class Modulation:
         NATIVE_METHODDS.autd3capi.AUTDModulationSinePressure(byref(mod.modulation_ptr), freq, amp, offset)
         return mod
 
-    @staticmethod
+    @ staticmethod
+    def sine_legacy(freq: float, amp=1.0, offset=0.5):
+        mod = Modulation()
+        NATIVE_METHODDS.autd3capi.AUTDModulationSineLegacy(byref(mod.modulation_ptr), freq, amp, offset)
+        return mod
+
+    @ staticmethod
     def square(freq: int, low: int = 0, high: int = 255):
         mod = Modulation()
         NATIVE_METHODDS.autd3capi.AUTDModulationSquare(byref(mod.modulation_ptr), freq, low, high)
@@ -261,42 +337,6 @@ class Sequence:
     def __del__(self):
         NATIVE_METHODDS.autd3capi.AUTDDeleteSequence(self.seq_ptr)
 
-    @staticmethod
-    def sequence():
-        seq = Sequence()
-        NATIVE_METHODDS.autd3capi.AUTDSequence(byref(seq.seq_ptr))
-        return seq
-
-    @staticmethod
-    def circum(center, normal, radius, num_points):
-        seq = Sequence()
-        NATIVE_METHODDS.autd3capi.AUTDCircumSequence(
-            byref(
-                seq.seq_ptr),
-            center[0],
-            center[1],
-            center[2],
-            normal[0],
-            normal[1],
-            normal[2],
-            radius,
-            num_points)
-        return seq
-
-    def add_point(self, point):
-        NATIVE_METHODDS.autd3capi.AUTDSequenceAddPoint(self.seq_ptr, point[0], point[1], point[2])
-
-    def add_points(self, points):
-        size = len(points)
-        points_array = np.zeros([size * 3]).astype(np.float64)
-        for i, p in enumerate(points):
-            points_array[3 * i] = p[0]
-            points_array[3 * i + 1] = p[1]
-            points_array[3 * i + 2] = p[2]
-        points_array = np.ctypeslib.as_ctypes(points_array)
-
-        NATIVE_METHODDS.autd3capi.AUTDSequenceAddPoints(self.seq_ptr, points_array, size)
-
     def set_frequency(self, freq: float):
         return NATIVE_METHODDS.autd3capi.AUTDSequenceSetFreq(self.seq_ptr, freq)
 
@@ -316,11 +356,68 @@ class Sequence:
         return NATIVE_METHODDS.autd3capi.AUTDSequenceSamplingPeriod(self.seq_ptr)
 
 
+class PointSequence(Sequence):
+    def __init__(self):
+        super().__init__()
+        NATIVE_METHODDS.autd3capi.AUTDSequence(byref(self.seq_ptr))
+
+    def __del__(self):
+        super().__del__()
+
+    @ staticmethod
+    def circum(center, normal, radius, num_points):
+        seq = PointSequence()
+        NATIVE_METHODDS.autd3capi.AUTDCircumSequence(
+            byref(
+                seq.seq_ptr),
+            center[0],
+            center[1],
+            center[2],
+            normal[0],
+            normal[1],
+            normal[2],
+            radius,
+            num_points)
+        return seq
+
+    def add_point(self, point, duty: int = 0xFF):
+        return NATIVE_METHODDS.autd3capi.AUTDSequenceAddPoint(self.seq_ptr, point[0], point[1], point[2], duty)
+
+    def add_points(self, points, duties):
+        points_size = len(points)
+        points_array = np.zeros([points_size * 3]).astype(np.float64)
+        for i, p in enumerate(points):
+            points_array[3 * i] = p[0]
+            points_array[3 * i + 1] = p[1]
+            points_array[3 * i + 2] = p[2]
+        points_array = np.ctypeslib.as_ctypes(points_array)
+
+        duties_size = len(duties)
+        duties_array = np.zeros([duties_size]).astype(np.uint8)
+        for i, d in enumerate(duties):
+            duties_array[i] = d
+        duties_array = np.ctypeslib.as_ctypes(duties_array)
+
+        return NATIVE_METHODDS.autd3capi.AUTDSequenceAddPoints(self.seq_ptr, points_array, points_size, duties_array, duties_size)
+
+
+class GainSequence(Sequence):
+    def __init__(self, gain_mode: GainMode = GainMode.DUTY_PHASE_FULL):
+        super().__init__()
+        NATIVE_METHODDS.autd3capi.AUTDGainSequence(byref(self.seq_ptr), gain_mode)
+
+    def __del__(self):
+        super().__del__()
+
+    def add_gain(self, gain: Gain):
+        return NATIVE_METHODDS.autd3capi.AUTDSequenceAddGain(self.seq_ptr, gain.gain_ptr)
+
+
 class Link:
     def __init__(self):
         self.link_ptr = c_void_p()
 
-    @staticmethod
+    @ staticmethod
     def enumerate_adapters():
         NATIVE_METHODDS.init_autd3capi_soem_link()
         res = []
@@ -337,20 +434,32 @@ class Link:
 
         return res
 
-    @staticmethod
-    def soem(ifname, dev_num, cycle_ticks=1):
+    @ staticmethod
+    def soem(ifname, dev_num, cycle_ticks=1, error_handler=None):
         NATIVE_METHODDS.init_autd3capi_soem_link()
         link = Link()
-        NATIVE_METHODDS.autd3capi_soem_link.AUTDLinkSOEM(byref(link.link_ptr), ifname.encode('utf-8'), dev_num, cycle_ticks)
+        error_handler = ErrorHandlerFunc(error_handler) if error_handler is not None else None
+        NATIVE_METHODDS.autd3capi_soem_link.AUTDLinkSOEM(byref(link.link_ptr), ifname.encode('utf-8'), dev_num, cycle_ticks, error_handler)
         return link
 
-    @staticmethod
+    @ staticmethod
     def twincat():
         link = Link()
         NATIVE_METHODDS.autd3capi_twincat_link.AUTDLinkTwinCAT(byref(link.link_ptr))
         return link
 
-    @staticmethod
+    @ staticmethod
+    def remote_twincat(remote_ip_addr, remote_ams_net_id, local_ams_net_id):
+        link = Link()
+        NATIVE_METHODDS.autd3capi_twincat_link.AUTDLinkRemoteTwinCAT(
+            byref(
+                link.link_ptr),
+            remote_ip_addr.encode('utf-8'),
+            remote_ams_net_id.encode('utf-8'),
+            local_ams_net_id.encode('utf-8'))
+        return link
+
+    @ staticmethod
     def emulator(port, autd):
         link = Link()
         NATIVE_METHODDS.autd3capi_emulator_link.AUTDLinkEmulator(byref(link.link_ptr), port, autd.p_cnt)
@@ -423,51 +532,51 @@ class AUTD:
     def _free(self):
         NATIVE_METHODDS.autd3capi.AUTDFreeController(self.p_cnt)
 
-    @property
+    @ property
     def is_open(self):
         return NATIVE_METHODDS.autd3capi.AUTDIsOpen(self.p_cnt)
 
-    @property
+    @ property
     def silent_mode(self):
         return NATIVE_METHODDS.autd3capi.AUTDIsSilentMode(self.p_cnt)
 
-    @silent_mode.setter
+    @ silent_mode.setter
     def silent_mode(self, value: bool):
         return NATIVE_METHODDS.autd3capi.AUTDSetSilentMode(self.p_cnt, value)
 
-    @property
+    @ property
     def force_fan(self):
         return NATIVE_METHODDS.autd3capi.AUTDIsForceFan(self.p_cnt)
 
-    @force_fan.setter
+    @ force_fan.setter
     def force_fan(self, value: bool):
         return NATIVE_METHODDS.autd3capi.AUTDSetForceFan(self.p_cnt, value)
 
-    @property
+    @ property
     def wavelength(self):
         return NATIVE_METHODDS.autd3capi.AUTDGetWavelength(self.p_cnt)
 
-    @wavelength.setter
+    @ wavelength.setter
     def wavelength(self, wavelength: float):
         NATIVE_METHODDS.autd3capi.AUTDSetWavelength(self.p_cnt, wavelength)
 
-    @property
+    @ property
     def attenuation(self):
         return NATIVE_METHODDS.autd3capi.AUTDGetAttenuation(self.p_cnt)
 
-    @attenuation.setter
+    @ attenuation.setter
     def attenuation(self, attenuation: float):
         NATIVE_METHODDS.autd3capi.AUTDSetAttenuation(self.p_cnt, attenuation)
 
-    @property
+    @ property
     def reads_fpga_info(self):
         NATIVE_METHODDS.autd3capi.AUTDIsReadsFPGAInfo(self.p_cnt)
 
-    @reads_fpga_info.setter
+    @ reads_fpga_info.setter
     def reads_fpga_info(self, value: bool):
         NATIVE_METHODDS.autd3capi.AUTDSetReadsFPGAInfo(self.p_cnt, value)
 
-    @property
+    @ property
     def fpga_info(self):
         infos = np.zeros([self.num_devices()]).astype(np.ubyte)
         pinfos = np.ctypeslib.as_ctypes(infos)
@@ -508,18 +617,21 @@ class AUTD:
     def num_transducers(self):
         return NATIVE_METHODDS.autd3capi.AUTDNumTransducers(self.p_cnt)
 
-    def send(self, gain: Gain, mod: Modulation):
-        return NATIVE_METHODDS.autd3capi.AUTDSendGainModulation(self.p_cnt, gain.gain_ptr, mod.modulation_ptr)
+    def send(self, gain: Gain, mod: Modulation, wait_for_msg_processed: bool = True):
+        return NATIVE_METHODDS.autd3capi.AUTDSendGainModulation(self.p_cnt, gain.gain_ptr, mod.modulation_ptr, wait_for_msg_processed)
 
-    def send_gain(self, gain: Gain):
-        return NATIVE_METHODDS.autd3capi.AUTDSendGain(self.p_cnt, gain.gain_ptr)
+    def send_gain(self, gain: Gain, wait_for_msg_processed: bool = True):
+        return NATIVE_METHODDS.autd3capi.AUTDSendGain(self.p_cnt, gain.gain_ptr, wait_for_msg_processed)
 
     def send_modulation(self, mod: Modulation):
         return NATIVE_METHODDS.autd3capi.AUTDSendModulation(self.p_cnt, mod.modulation_ptr)
 
-    def send_seq(self, seq: Sequence):
+    def send_seq(self, seq: PointSequence):
         return NATIVE_METHODDS.autd3capi.AUTDSendSequence(self.p_cnt, seq.seq_ptr)
 
+    def send_gain_seq(self, seq: GainSequence):
+        return NATIVE_METHODDS.autd3capi.AUTDSendGainSequence(self.p_cnt, seq.seq_ptr)
+
     def stm(self):
         handle = c_void_p()
         NATIVE_METHODDS.autd3capi.AUTDSTMController(byref(handle), self.p_cnt)
